/* ipofix.c - a program for closing timing.  By Jeff Winston  
Rev. 1.1, 1/14/00  Copyright (C) 1999, 2000

IPOFIX is a program used to fix timing problems.  It performs the
following functions:

1) Given the report files from the Synopsys IPO functions, it applies
the upsize changes listed in the report, and renames the gates (which
forces Primetime to re-calculate delays based on R C loads and/or 
wireload models).  Note that the report file format used by Synopsys
is simple enough to be re-created by hand for applying user-specified
changes, or changes generated by some other method. 

2) Given a worst-case Primetime report file, IPOFIX upsizes and
renames gates whose delays exceed user-specified values.

3) Given a worst-case Primetime report file, or a transition report
file generated by VLSI's DCALC tool, IPOFIX attempts to fix transition
violations by upsizing gates.  (Upsized gates are renamed as above). 

4) Given a min-delay Primetime report file, or an explicit list,
IPOFIX inserts hold buffers where needed.

Ipofix can also write out and read back in its changes database for
external processing.

See the accompanying help file for detailed instructions

---------------------------------------------------------------------
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.  or check out http://www.gnu.org/copyleft/gpl.html
---------------------------------------------------------------------
*/

/* All the defines below should be adjusted up per project if neeeded.  Also, certain sections of code are
  project specific and should be changed for different projects.  These are all marked by the words "PROJECT-SPECIFIC" 
  other sections are specific to the ASIC library naming used and are marked by the words LIBRARY-SPECIFIC 
  note that for libraries that do not denote strength in their last character this code may need some significant changes 
*/

# include  <stdio.h>
# include  <string.h>

# define LOGIC_TOP  "ucore"       /* PROJECT-SPECIFIC: top level below chip level - no fixes above this level */
# define MAX_HIERARCHY      20    /* max levels of hierarchy in an hierarchical instance name */
# define MAXTOKENS       10000    /* max number of tokens in a verilog statement */
# define MAX_MODULES      1000    /* maximum number of modules in the input file */
# define MAX_MOD_CALLS    5000    /* maximum number of module calls to other non-primitive modules in the input file */
# define MAX_CHANGES     50000    /* maximum number of changes in the changes file, including duplicates */
# define MAX_TRANSITIONS 10000    /* max size of transitions database, used to remove duplications */

char 
instr[255],		/* general string used for reading files */
fullstr[655350],	/* used for holding an entire verilog statement */
tmpstr[80], 
instname[80],		/* temporary to hold instance name */
modtmp[80],		/* temps to hold module names */
modname[80], 
hier_instname[132],	/* temporary to hold hierarchical instance name */
oldgate[32],		/* temporary to hold old gate name */
bc_sdf_dly_r[10],		/* holds default SDF delays for hold buffers */
bc_sdf_dly_f[10],
wc_sdf_dly_r[10],
wc_sdf_dly_f[10],

tokens[MAXTOKENS][80],	    /* returned tokens from a verilog statement */
itokens[MAX_HIERARCHY][80], /* returns tokens from parsing a hierarcheral instance name */

lines[1000][255],	    /* temporary to hold lines making up a verilog statement */


i_mods[MAX_MODULES][48];    /* list of all module names in the input file */
long modcnt;		    /* tracks current array size */

char		            /* the module reference data base. */
i_mmod[MAX_MOD_CALLS][48],  /* module name */
i_cmod[MAX_MOD_CALLS][48],  /* called module name */
i_inst[MAX_MOD_CALLS][48];  /* instance name of called module */
long i_cnt;		    /* tracks current array size */


char			      /* list of all recorded changes */
c_mod[MAX_CHANGES][48],       /* change is in this module */
c_inst[MAX_CHANGES][48],      /* of this instance */
c_old[MAX_CHANGES][48],       /* this is the old gate */
c_new[MAX_CHANGES][48],       /* this is new gate  */
c_hiername[MAX_CHANGES][80],  /* this is the hierarchical instance name */
c_pin[MAX_CHANGES][5];        /* this is the pin that gets a buffer */
long			      
c_used[MAX_CHANGES],	      /* marked whether the change has or should be applied */
c_exact[MAX_CHANGES],         /* marks whether the old gate is an exact match */
c_trans[MAX_CHANGES],         /* marks whether it's a change due to transition time */
c_buffers[MAX_CHANGES];       /* the number of hold buffers to add */
long c_cnt;		      /* tracks current array size */

long t_type[MAX_TRANSITIONS];      /* database used for removing duplicate transitions */
char t_name[MAX_TRANSITIONS][80];
long t_cnt;			   /* tracks current array size */


/* filenames */
char  infile[80],  outfile[80],  chfile[80],  rptfile[80],  dbfile[80],  trptfile[80],  bsdf_file[80], wsdf_file[80];  

/* file pointers */
FILE *infile_ptr, *outfile_ptr, *chfile_ptr, *rptfile_ptr, *dbfile_ptr, *trptfile_ptr, *bsdf_ptr,      *wsdf_ptr;

# define NOTYPE 0		/* convenience flag for determining filetype */
# define PRIMETIME 1
# define PRIMEHOLD 2
# define PRIMETRANS 3
# define SYNIPO 4
# define HOLDLIST 5
# define CHANGEDB 6

# define TRANS_CANT_FIX 1	/* convenience flags for transition disposition type */
# define TRANS_SKIP     2
# define TRANS_NONEED   3
# define TRANS_FIXED    4


long		
		    /* counters */
    trans_cnt,	      /* counts transition changes */
    buffer_cnt,	      /* counts number of hold buffers inserted */
    soscnt,	      /* counts transition problems on /so pins */
    nodenum,	      /* nodenum used for hold buffers */
    actual_chgcnt,    /* counts actual changes */
    total_chgcnt,     /* counts total changes */
    infile_line_cnt,  /* counts lines in input file */
    chgcnt,	      /* counts changes within a module */
    argcnt,	      /* counts comment line arguments */
 
		   /* flags */
    write_changes_db, /* flag - write changes db and exit when set */
    gen_db_and_exit,  /* flag - generate database and exit when set */
    diag,	      /* diags enabled flag */
    open_sdf,	      /* are sdf files open? */
    trpfile_open,     /* transition report file is open */
    active_mod,	      /* 1= we processing an active module */
    db_file,	      /* 1 = there is a database file */



		   /* for parsing */
    ilevel,	      /* current level of hierarchy in itokens */
    rpt_line,	      /* current line of report file */
    ntk,	      /* number of tokens returned in  tokens[] */
    intk,	      /* number of tokens returned in itokens[] */
    all_done,	      /* used for parsing verilog lines */
    parsing_comment,  /* used for tracking comment parsing */
    start_line,	      /* which line in stored input-lines buffer is first to be output */
    inst_tok,	      /* the token # of the instance */
    pin_tok,	      /* the token # of the pin */
    tok,	      /* used to walk through tokens on an input line */
    chgnum,	      /* used to walk through changes database */

		  /* misc */
    drive,	      /* drive strength of buffer */
    max_drive,	      /* max drive strength allowed */
    filetype,	      /* which type of report file are we processing */
    x, y,	      /* generic loop variables */
    done,	      /* general done variable */
    percent_disp,     /* used for displaying doneness of database check */
    percent_incr, 
    percent_done, 
    x1,y1;	       /* temporaries - used in 1 place */

float 
      slack,		/* slack from PT output */	
      delay,		/* used to calculate buffer delay */
      added_delay, 
      trans,		/* transition value */
      bfdly,		/* single buffer delay value */
      lim,		/* limits from command line */
      lo_lim, 
      mid_lim, 
      hi_lim, 
      lim1,
      lim2,
      old,		/* temps */
      new, 
      f1, 
      f2;

main(argc,argv)
int argc;                             /* Parse input file from command line */
char *argv[];
{
    parsing_comment=0;
    argcnt=1;	/* counts command line arguments */

    ck_argv(argc);		/* check for enough arguments.  exit with error message if not */
    ++argv;			/* get qualifiers, if there are any */

    strcpy(tmpstr,*argv);

    gen_db_and_exit=0;	/* flags set from command line */
    write_changes_db = 0;
    diag=0; 
    db_file=0;

    if (strcmp(tmpstr,"-t")==0) 
    {
	gen_db_and_exit=1;
	ck_argv(argc);
	++argv;
    }
    else if (strcmp(tmpstr,"-c")==0) 
    {
	write_changes_db = 1;
	ck_argv(argc);
	++argv;
    }
    else if (strcmp(tmpstr,"-d")==0) 
    {
	diag=1;
	printf("Diagnostics enabled\n");
	ck_argv(argc);
	++argv;
    }
    else if (tmpstr[0] == '-') 
    {
	printf(" Unrecognized qualifier %s.  Exiting\n",tmpstr);
	exit(1);
    }
    			/* get input filename, try to open file */
    strcpy(infile,*argv);		
    if ( (infile_ptr=fopen(infile, "r")) ==NULL )
    {
            printf("Error:  Error opening input file /%s/\n\n",infile);
            exit(1);
    }

    if (gen_db_and_exit == 0)
    {
	ck_argv(argc);
        ++argv;			/* get report filename, try to open file  */
        strcpy(rptfile,*argv);		
        if ( (rptfile_ptr=fopen(rptfile, "r")) ==NULL )
        {
            printf("Error:  Error opening report file /%s/\n\n",rptfile);
            exit(1);
        }
    }
    ck_argv(argc);
    ++argv;				
    strcpy(outfile,*argv);     /* get output filename, try to open file   */
    if ((outfile_ptr=fopen(outfile, "w")) ==NULL)
    {
        printf("Error:  Error opening output file /%s/\n\n",outfile);
        exit(1);
    }

    if ((gen_db_and_exit == 0) && (argc > argcnt))
    {			       /* get database filename, try to open file  */
	argcnt++;
        ++argv;
	db_file=1;
	strcpy(dbfile,*argv);
        if ( (dbfile_ptr=fopen(dbfile, "r")) ==NULL )
        {
            printf("Error:  Error opening input file /%s/\n\n",dbfile);
            exit(1);
        }
    }
        
    if (argc > argcnt) ck_argv(0);


    strcpy(trptfile,"trans_report");  /* this filename, if used, is fixed */
    strcpy(wsdf_file,"wc.sdf");	      /* this filename, if used, is fixed */
    strcpy(bsdf_file,"bc.sdf");	      /* this filename, if used, is fixed */
    

/* Read in reference database */

    if (gen_db_and_exit==1) /* build new reference table and exit */
    {
	build_ref_table();

        for (x=0; x<i_cnt;x++)
        {
	    fprintf(outfile_ptr,"%s %s %s\n",i_mmod[x],i_cmod[x],i_inst[x]);
        }
        printf("%d module references.\n",i_cnt);
        printf("%s written\n",outfile);
        exit(1);
     }
     else if ((gen_db_and_exit==0) && (db_file==0))  /* no database file - build reference table */
     {
	build_ref_table();
        fopen(infile, "r");
     }
     else read_reference_table();	/* read reference table in */


/* Reading in report file */

    printf("Reading in report file...\n");

    /* zero a bunch of globals */
    c_cnt=0; 
    trpfile_open=0; 
    soscnt=0;
    nodenum=0;
    trans_cnt=0;
    open_sdf=0;
    filetype=0;
    t_cnt=0;

    while (get_rpt_line(1) != 0)    /* get next line from report file */
    {
       ntk = get_par_tokens(instr,0); /* parse out ( and ) */
       if (ntk > 1) 
       {

		/* identify file type */

	       if (find_string(instr,"$ END_CHANGES_FILE")>0) filetype = NOTYPE;

	  else if (find_string(instr,"START_CHANGES_DATABASE_INPUT")>0)
	  {
		if (ntk != 7)
		{
			printf("Error: Bad number of tokens on DATABASE_INPUT line (should be 7)\n");
			exit(1);
		}				      

	        printf("Reading changes in database format...\n");
		filetype = CHANGEDB;
		strcpy(bc_sdf_dly_r,tokens[2]);
		strcpy(bc_sdf_dly_f,tokens[3]);
		strcpy(wc_sdf_dly_r,tokens[4]);
		strcpy(wc_sdf_dly_f,tokens[5]);
		sscanf(tokens[6],"%d",&nodenum);


	  }
	  else if (find_string(instr,"$ MODIFIED_CELLS")>0)  /* SYNIPO file type flag */
	  {
		filetype = SYNIPO;
		get_rpt_line(1);
	  }
	  else if ((filetype == NOTYPE)  && (find_string(instr,"$ Design:")>0))  /* get design name if SYNIPO file */
	  {
	  	  strcpy(modname,tokens[2]);
		  printf("Reading changes for design %s\n",modname);
	  }
	  else if (find_string(instr,"$ Primetime")>0) /* primetime delay violations file */
	  {
		filetype = PRIMETIME;
		if (ntk != 6)
		{
			printf("Error in report file.  First line of Primetime block must be\n");
			printf("$ Primetime <top module> <threshold 1> <threshold 2> <threshold 3>\n");
			exit (1);
		}
		strcpy(modname,tokens[2]);
		sscanf(tokens[3],"%f",&lo_lim);
		sscanf(tokens[4],"%f",&mid_lim);
		sscanf(tokens[5],"%f",&hi_lim);
		  printf("Reading changes out of Primetime output.  Top module: %s.  Limits: %f %f %f\n",modname,lo_lim,mid_lim,hi_lim);
	  }
	  else if ((find_string(instr,"$ Primehold")>0) || (find_string(instr,"$ Holdvio")>0)) /* primetime hold vios or list */
	  {
		filetype = PRIMEHOLD;
		if (find_string(instr,"$ Holdvio")>0) filetype = HOLDLIST;
		open_sdf=2; /* schedule opening of sdf files */

		if ((nodenum > 0) && (ntk != 6))
		{
			printf("Error in report file.  First line of Primetime  hold block must be\n");
			printf("$ Primehold <top module> <upper threshold> <lower threshold> <buffer delay>\n");
			exit (1);
		}
		else if ((nodenum ==0) && (ntk != 7) && (ntk != 11))
		{
			printf("Error in report file.  First line of Primetime hold block must be\n");
			printf("$ Primehold <top module> <upper threshold> <lower threshold> <buffer delay> <4-digit seed> \\ \n");
			printf("[<BC sdf rise dly>] [<BC sdf fall dly>] [<WC sdf rise dly>] [<WC sdf fall dly>]\n");
			exit (1);
		}
		strcpy(modname,tokens[2]);	/* parse out command line */
		sscanf(tokens[3],"%f",&hi_lim);
		sscanf(tokens[4],"%f",&lo_lim);
		sscanf(tokens[5],"%f",&bfdly);
		if (ntk==7) 
		{
			sscanf(tokens[6],"%d",&nodenum);
			strcpy(bc_sdf_dly_r,"0.062");	/* defaults for 0.2u */
			strcpy(bc_sdf_dly_f,"0.063");
			strcpy(wc_sdf_dly_r,"0.136");
			strcpy(wc_sdf_dly_f,"0.129");
		}
		if (ntk==11)
		{
			strcpy(bc_sdf_dly_r,tokens[7]);
			strcpy(bc_sdf_dly_f,tokens[8]);
			strcpy(wc_sdf_dly_r,tokens[9]);
			strcpy(wc_sdf_dly_f,tokens[10]);
		}
			/* print info */
		printf("Reading hold violations out of Primetime output.\n");
		printf("Top module: %s.  Limits: %f <= fix <= %f bufdly = %f\n",
		       modname,lo_lim,hi_lim,bfdly);
		printf("SDF Values:  BC: Rise: %s Fall: %s. WC: Rise: %s Fall: %s.\n",
		bc_sdf_dly_r, bc_sdf_dly_f, wc_sdf_dly_r, wc_sdf_dly_f);
		printf("Additions of groups larger than 3 buffers will be listed below\n");
	  }
		/* transition data */	  
	  else if (find_string(instr,"$ Transitions")>0)  /* if primetime transitions file */
	  {
		filetype = PRIMETRANS;			  /* set file type flag */
		x=0;
		while (strcmp(tokens[x],"Transitions")!=0)  /* check for valid number of tokens on line */
		{
			x++;
			if (x == ntk)
			{
				printf("Bad line %d: %s",rpt_line,instr);
				exit(0);
			}	
		}
		if (x+3 >= ntk)
		{
			printf("Bad line %s",instr);
			exit(0);
		}	
		strcpy(modname,tokens[x+1]);		/* get module name and limits off of line */
		sscanf(tokens[x+2],"%f",&lim1);
		sscanf(tokens[x+3],"%f",&lim2);
		if (ntk == (x+4))
		{
		   lo_lim=1; hi_lim=2;
		}
		else if (ntk == (x+6))
		{
			sscanf(tokens[x+4],"%f",&lo_lim);
			sscanf(tokens[x+5],"%f",&hi_lim);
		}
		else
		{
			printf("Bad number of tokens on line %s\n",instr);
			exit(1);
		}
		printf("Reading changes out of Transitions output.  Top module: %s.\nTransition Limits: %f %f, Increments: %f %f\n"
				,modname,lim1, lim2, lo_lim, hi_lim);
          }

			/* process transition-related data */
	  else if (((filetype == PRIMETRANS) && ((strcmp(tokens[1],"TRANSDATA")==0) || (strcmp(tokens[1],"FORTRANSFIX")==0))) ||
		   ((ntk==9) && ( (strcmp(tokens[0],"RISE")==0) || (strcmp(tokens[0],"FALL")==0)  ) &&
		    ( (strcmp(tokens[1],"TRUE")==0) || (strcmp(tokens[1],"FALSE")==0) ) ) ) 
	  {
	        	/* if file is from DCALC, parse line for (old)gate transition time and hier_instname */
		if ((ntk==9) && ( (strcmp(tokens[0],"RISE")==0) || (strcmp(tokens[0],"FALL")==0)  ) &&
		    ( (strcmp(tokens[1],"TRUE")==0) || (strcmp(tokens[1],"FALSE")==0) ) )

		{
			strcpy(oldgate,tokens[7]);
			sscanf(tokens[2],"%f",&trans);
			sscanf(tokens[3],"%f",&f1);
			sscanf(tokens[4],"%f",&f2);		/* PROJECT-SPECIFIC top level module is assumed */
			if (f2 > trans) trans = f2;		/* to be either chip or phoenix */
			if (f1 > trans) trans = f1;
			     if (find_string(tokens[8],"chip/")==1)    strcpy(hier_instname,&tokens[8][5]);
			else if (find_string(tokens[8],"phoenix/")==1) strcpy(hier_instname,&tokens[8][8]);
			else                                           strcpy(hier_instname,tokens[8]);		

		}
		/* otherwise parse primetime output (generated using trans.c */
		else if (strcmp(tokens[1],"TRANSDATA")==0)	/* get same info from TRANSDATA line */
		{
			strcpy(hier_instname,tokens[2]);		
			strcpy(oldgate,tokens[3]);
			sscanf(tokens[4],"%f",&trans);
		}
		else
		{
								/* parse FORTRANSFIX section */
			sscanf(tokens[2],"%f",&trans);
			while (find_string(instr,"Driver")!=1) get_rpt_line(1);
			get_rpt_line(2);
			ntk = get_tokens(instr);
			strcpy(hier_instname,tokens[0]);
			get_rpt_line(3);
			ntk = get_tokens(instr);
			if (strcmp(tokens[0],"Load")!=0)
			{
				printf("Bad line %d, looking for Load %s\n",rpt_line,instr);
				exit(0);
			}
			get_rpt_line(2);
			ntk = get_tokens(instr);
			if (ntk < 3)
			{
				printf("Bad line %d, need 3 tokens %s\n",rpt_line,instr);
				exit(0);
			}
			strcpy(oldgate,tokens[2]);
		}

		/* at this point, we have collected transition info from any of the 3 acceptable input methods - now we process it */
		if (diag >0) printf("Processing %s\n",hier_instname);

		lim = lim1;			     /* set transition threshold   PROJECT-SPECIFIC */
						     /* this code should assign the lower lim2 to hierarchical instance
						        names fitting a desired pattern */

		if ((find_string(hier_instname,"/uu")>0) ||(find_string(hier_instname,"/bistrouter")>0)) lim=lim2;

		strcpy(tmpstr,&oldgate[strlen(oldgate)-1]);
		sscanf(tmpstr,"%d",&drive);		/* get current drive from gate */
							   
								/* LIBRARY-SPECIFIC max_drive varies by gate type */
		max_drive=3;				    /* set max drive */
		if ((find_string(oldgate,"in01d")>0) || (find_string(oldgate,"ni01d")>0) ) max_drive=4;
		else if (find_string(oldgate,"ad01d")>0) max_drive=2;


		 if (drive >= max_drive) /* can't fix it */
		 {
			max_drive=0;
			trans_entry(TRANS_CANT_FIX, hier_instname, oldgate, trans);
		 }
		 else if (trans <= lim)       /* don't need to fix it */
		 {
			max_drive=0;
			trans_entry(TRANS_NONEED, hier_instname, oldgate, trans);
		 }	
		 else if (								/* LIBRARY/VENDOR-SPECIFIC */
		      (find_string(hier_instname,"/so")==strlen(hier_instname)-2) ||	/* cant upsize /so's */
		      (find_string(hier_instname,"upad/cin")>0) ||			/* cant upsize input pads */
		      ((find_string(hier_instname,LOGIC_TOP)!=1) &&	/* if not in the core */
		       (find_string(hier_instname,"uu")!=1) &&	        /* don't touch uu, oen, or tap controller */
		       (find_string(hier_instname,"oen")==0) &&		/* (these 3 names used by VTI tools) */
		       (find_string(hier_instname,"utap")==0) ) ||

		      (find_string(oldgate,"ram")>0) ||		/* these 2 are PROJECT-SPECIFIC */	
		      (find_string(hier_instname,"spg")>0))	/* they specify characteristics of instance names that
								   should not have transition errors fixed */
		 {
			max_drive=0;
			trans_entry(TRANS_SKIP, hier_instname, oldgate, trans);	
		 }

		 else   /* if we can improve it */
		 {
			get_inst_and_pin(modname,hier_instname,0);   /* get instance and pin name */
			drive++;				   /* pick the improvement */
			if (trans > lim+lo_lim) drive++;
			if (trans > lim+hi_lim) drive++;
			if (drive > max_drive) drive=max_drive;
		        strncpy(c_new[c_cnt],&oldgate[0],strlen(oldgate)-1);
                        sprintf(tmpstr,"%d",drive);
		        strcat(c_new[c_cnt],tmpstr);
			if (find_string(itokens[inst_tok],"_XQX")>0)	/* strip _XQX */
			{
				strncpy(tmpstr,&itokens[inst_tok][0],strlen(itokens[inst_tok])-4);
				strncpy(&tmpstr[strlen(itokens[inst_tok])-4],"\0",1);
				strcpy(c_inst[c_cnt],tmpstr);
			}						/* enter it in changes data base */
			else strcpy(c_inst[c_cnt],itokens[inst_tok]);
			strcpy(c_old[c_cnt],oldgate);			
			strcpy(c_mod[c_cnt],modtmp);
			c_used[c_cnt]=0; 
			c_exact[c_cnt]=0;
			c_trans[c_cnt]=1;
			if (diag>0)  printf("Recording change... mod: %s  inst: %s  old: %s  new: %s\n",
                              c_mod[c_cnt],c_inst[c_cnt],c_old[c_cnt],c_new[c_cnt]);
			c_cnt++;
			trans_cnt++;
			limit_check(c_cnt,MAX_CHANGES,"changes",instr);
			open_trans_file();
			if (diag>0) trans_entry(TRANS_FIXED, hier_instname, oldgate, trans);
		  }
	  }
	  /* process hold list data */

	  else if ((filetype == HOLDLIST) && (find_string(instr,LOGIC_TOP)>0) && (find_string(instr,"//")==0))
	  {
		ntk = get_tokens(instr);
		if (ntk != 2) 
		{
			printf("Incorrect # of tokens: %s\n",instr);
			exit(1);
		}
		get_inst_and_pin(modname,tokens[0],0);   /* get instance and pin name */
		put_hiername();
		sscanf(tokens[1],"%d",&y);			/* enter in changes database */
		strcpy(c_inst[c_cnt],itokens[inst_tok]);			
		strcpy(c_old[c_cnt],"null");
		strcpy(c_new[c_cnt],"null");
		strcpy(c_mod[c_cnt],modtmp);
		sprintf(c_pin[c_cnt],".%s",itokens[pin_tok]);	/* LIBRARY-SPECIFIC Pin names */
		if ((strcmp(itokens[pin_tok],"d" )!=0) && 
		    (strcmp(itokens[pin_tok],"da")!=0) && 
                    (strcmp(itokens[pin_tok],"wen")!=0) &&
		    (strcmp(itokens[pin_tok],"si")!=0)  )
		{
			printf("Illegal pin on instance %s\n",tokens[0]);
			exit(1);
		}
		c_used[c_cnt]=0; 
		c_exact[c_cnt]=-1;
		c_trans[c_cnt]=0;
		c_buffers[c_cnt]=y;
		if ((diag>0) || (y>3) ) 
		   printf("Recording change... mod: %s  inst: %s  pin: %s  buffers: %d\n",
                              c_mod[c_cnt],c_inst[c_cnt],c_pin[c_cnt],c_buffers[c_cnt]);
	        c_cnt++;
		limit_check(c_cnt,MAX_CHANGES,"changes",instr);
	  }

	  /* process primetime hold data */
	  else if ((filetype == PRIMEHOLD) && (find_string(instr,"Startpoint:")>0))	/* parse through the paragraph */
	  {
	      strcpy(tmpstr,tokens[1]);
	      get_rpt_line(1);
	      x=0;
	      while ((x<3) && (find_string(instr,"Endpoint:")==0)) {get_rpt_line(1); x++;}
	      if (x==3)
	      {
		printf("Couldn't find Endpoint: heading in section starting with Startpoint: %s, line %d\n",tmpstr,rpt_line);
		exit(1);
              }
	      ntk = get_tokens(instr);
	      if (strcmp(tokens[1],tmpstr)==0)	/* skip hold vios to same flop */
	      {
		ntk=0;
		printf("Skipping hold violation from/to flop %s\n",tokens[1]);
	      }
	      else
	      {
		  get_rpt_line(1);
		  while (find_string(instr,"------------")==0) get_rpt_line(1);
		  x=0;
		  while (find_string(instr,tokens[1])==0) {get_rpt_line(1); x++;}
		  if (x >22)
		  {
			printf("Error: Parsed too many lines, couldn't find flop for: %s, line %d\n",tokens[1],rpt_line);
			exit(1);
		  }
		  ntk = get_par_tokens(instr,0);
		  if (ntk == 2)
		  {
			get_rpt_line(1);
			ntk = get_par_tokens(instr,ntk);
		  }
	      }		
	      if (ntk >4)
	      {
		  if (find_string(tokens[0],"/SPF")>0)	/* PROJECT-SPECIFIC - skip hold violations for hierarchical 
							   instance names that have certain characteristics */
		  {
			printf("Skipping hold violation for %s\n",tokens[0]);
			slack=0;
		  }
		  else 
		  {
			get_inst_and_pin(modname,tokens[0],0);   /* get instance and pin name */
			strcpy(oldgate,tokens[1]);		
			while (strcmp(tokens[0],"slack")!=0)
			{
				if ((get_rpt_line(1) ==0) || (find_string(instr,"Startpoint")>0))
				{
					printf("Error: Problem parsing section starting with Endpoint: %s, line %d\n",oldgate,rpt_line);
					exit(1);
				}
				ntk = get_tokens(instr);
			}
			sscanf(tokens[ntk-1],"%f",&slack);
		  }
		  if (slack<0) slack = slack * -1; 
		  if ((hi_lim >= slack) && (slack >= lo_lim) && (find_string(instr,"VIOLATED")>0)) /* is there a hold vio? */
		  {
		     x=0;
		     added_delay=0;
		     while (added_delay < (slack+.001))	/* determine # of buffers needed */
		     {
			x++;
			added_delay = added_delay + bfdly;
	             }			
		     put_hiername();
		     strcpy(c_inst[c_cnt],itokens[inst_tok]);			
		     strcpy(c_old[c_cnt],oldgate);			
		     strcpy(c_mod[c_cnt],modtmp);
		     if ((strcmp(itokens[pin_tok],"d" )!=0)  &&		/* LIBRARY-SPECIFIC pin-names */
                         (strcmp(itokens[pin_tok],"da")!=0)  && 
                         (strcmp(itokens[pin_tok],"wen")!=0) &&
                         (strcmp(itokens[pin_tok],"si")!=0))
		     {
			printf("Illegal pin on instance %s/%s/%s\n",c_hiername[c_cnt],c_inst[c_cnt],itokens[pin_tok]);
			exit(1);
		     }
		     sprintf(c_pin[c_cnt],".%s",itokens[pin_tok]);	/* put change in database */
	             c_used[c_cnt]=0; 
		     c_exact[c_cnt]=1;
		     c_trans[c_cnt]=0;
		     c_buffers[c_cnt]=x;
		     if ((diag>0) || (x>3))
		           printf("Recording change... mod: %s  inst: %s  pin: %s  buffers: %d\n",
                              c_mod[c_cnt],c_inst[c_cnt],c_pin[c_cnt],c_buffers[c_cnt]);
	             c_cnt++;
		     limit_check(c_cnt,MAX_CHANGES,"changes",instr);
		  }
	      }
           }

	  /* process Primetime delay data */

	  else if ((filetype==PRIMETIME) && (		/* process lines with enough tokens, and LOGIC_TOP in the instance name */
			 (    (ntk>4)
			   && (find_string(instr,"/")>0)
			   && (find_string(instr,LOGIC_TOP)>0)) ||
			      ( (ntk == 2) && (find_string(tokens[0],LOGIC_TOP)==1) ) ) )	/* 2-line case */
          {
	      if (ntk == 2)	/* handle 2-line case */
	      {
		get_rpt_line(1);
		ntk = get_par_tokens(instr,ntk);
              }
	      if (ntk >4)	/* process tokens */
	      {
	         sscanf(tokens[2],"%f",&delay);
	         if (delay > lo_lim)
	         {
		    get_inst_and_pin(modname,tokens[0],0);   /* get instance and pin name */
		    max_drive=3;
		    if (find_string(tokens[1],"ram")>0) {drive=0;max_drive=0;} /* PROJECT-SPECIFIC - don't upsize instances
										  which have RAM in the instance name */
											
		    else								/* LIBRARY-SPECIFIC max drive */
		    {						/* set max size for gate */
			if ((find_string(tokens[1],"in01d")>0) || (find_string(tokens[1],"ni01d")>0) ) max_drive=4;
			if (find_string(tokens[1],"ad01d")>0) max_drive=2;
			strcpy(tmpstr,&tokens[1][strlen(tokens[1])-1]);
			sscanf(tmpstr,"%d",&drive);				
		    }
		    if (drive <max_drive) /* if room to upsize, do it */
		    {
			drive++;
			if (delay > mid_lim) drive++;
			if (delay > hi_lim) drive++;
			if (drive > max_drive) drive=max_drive;
		        strncpy(c_new[c_cnt],&tokens[1][0],strlen(tokens[1])-1);
                        sprintf(tmpstr,"%d",drive);
		        strcat(c_new[c_cnt],tmpstr);			/* enter in changes database */
			if (find_string(itokens[inst_tok],"_XQX")>0)
			{  
				strncpy(tmpstr,&itokens[inst_tok][0],strlen(itokens[inst_tok])-4);
				strncpy(&tmpstr[strlen(itokens[inst_tok])-4],"\0",1);
				strcpy(c_inst[c_cnt],tmpstr);
			}
			else strcpy(c_inst[c_cnt],itokens[inst_tok]);
			strcpy(c_old[c_cnt],tokens[1]);			
			strcpy(c_mod[c_cnt],modtmp);
			c_used[c_cnt]=0; 
			c_exact[c_cnt]=0;
			c_trans[c_cnt]=0;
			if (diag>0) printf("Recording change... mod: %s  inst: %s  old: %s  new: %s\n",
                              c_mod[c_cnt],c_inst[c_cnt],c_old[c_cnt],c_new[c_cnt]);
			c_cnt++;
			limit_check(c_cnt,MAX_CHANGES,"changes",instr);
		    }
	        }
	     }
	  }

	       /* process synopsys IPO data */
	  else if ((filetype==SYNIPO) && (ntk > 2))
	  {
		if ((strcmp(tokens[2],":")!=0) || (strcmp(tokens[3],"original")!=0) || (strcmp(tokens[4],"lib_cell")!=0))
		{
			printf ("Error: Unknown line in report file: %s\n",instr);
			exit(1);	  
		}
		strcpy(tmpstr,&tokens[1][strlen(tokens[1])-1]);
		sscanf(tmpstr,"%d",&new); 
		strcpy(tmpstr,&tokens[5][strlen(tokens[5])-1]);
		sscanf(tmpstr,"%d",&old); 

		if (old < new) /* only enter upsizings */
		{
			get_inst_and_pin(modname,tokens[0],1);   /* get instance and pin name */
			strcpy(c_mod[c_cnt],modtmp);
			strcpy(c_inst[c_cnt],itokens[inst_tok]);
			strcpy(c_old[c_cnt],tokens[5]);
			strcpy(c_new[c_cnt],tokens[1]);		
			c_used[c_cnt]=0;				/* enter in changes database */
			c_exact[c_cnt]=1;
			c_trans[c_cnt]=0;
			if (diag>0) printf("Recording change... mod: %s  inst: %s  old: %s  new: %s\n",c_mod[c_cnt],c_inst[c_cnt],c_old[c_cnt],c_new[c_cnt]);
			c_cnt++;
			limit_check(c_cnt,MAX_CHANGES,"changes",instr);
		}
	  }
	  /* read in database */
	  else if ((filetype == CHANGEDB) && (ntk == 9))
	  {
	    strcpy(c_mod[c_cnt],     tokens[0]);
	    strcpy(c_inst[c_cnt],    tokens[1]);
	    strcpy(c_old[c_cnt],     tokens[2]);
	    strcpy(c_new[c_cnt],     tokens[3]);
	    strcpy(c_hiername[c_cnt],tokens[4]);
	    strcpy(c_pin[c_cnt],     tokens[5]);
	    sscanf(tokens[6],"%d",&c_exact[c_cnt]);
	    sscanf(tokens[7],"%d",&c_trans[c_cnt]);
	    sscanf(tokens[8],"%d",&c_buffers[c_cnt]);
	    if (c_buffers[c_cnt] > 0) open_sdf=2;
	    c_used[c_cnt]=0;
	    c_cnt++;	  
	  }
	  else if ((filetype == CHANGEDB))
	  {
		printf("Error:  Incorrect number of tokens (should be 9) on line:\n%s\n",instr);
		exit(1);
	  }
	}
    }

   printf("Recorded %d changes, %d related to transitions (including duplicates).\n",c_cnt,trans_cnt);
    if ((trpfile_open>0) && (soscnt > 0)) fprintf(trptfile_ptr,"Also skipped %d flops (including duplicates) with transition violations on their .so output\n");

    if (diag>0)
    {
        for (x=0;x<c_cnt;x++)
        {
		printf("%s %s %s %s %s %s\n",c_mod[x],c_inst[x],c_old[x],c_new[x],c_hiername[x],c_pin[x]);
        }
    }


    /* Next step - reduce duplicates in database using sloppy n^2 search */

   printf("Resolving conflicts in changes database ...\n"); fflush(stdout);
   actual_chgcnt = c_cnt; 
   percent_disp = 10;		/* percent_* used for display only */
   percent_incr = c_cnt/10;
   percent_done = percent_incr;   

   for (x=0; x<c_cnt-1; x++) 
   {
       if (x > percent_done)
       { 
          printf("%2d percent done\n",percent_disp);
          percent_disp = percent_disp+10;
          percent_done = percent_done + percent_incr;
       }
       for (y=x+1;y<c_cnt; y++)
       {
	  if ((c_used[y]==0) && 
	      (strcmp(c_mod[x] ,c_mod[y] )==0) &&
	      (strcmp(c_inst[x],c_inst[y])==0) &&
	      (( c_buffers[x]==0) || (strcmp(c_pin[x],c_pin[y])==0)) )
	  {
	        if (diag>0)  printf("Multiple changes in module %s for instance %s\n",c_mod[x],c_inst[x]);
		if (c_buffers[x]>0)
		{
			if (c_buffers[y] > c_buffers[x]) c_buffers[x] = c_buffers[y];
		}
		else if (strcmp(c_new[x],c_new[y])!=0) 
		{
		        if (diag>0) printf("Conflicting changes for instance %s in module %s: Orig: %s, new: %s or %s, using ",
					   c_inst[x],c_mod[x],c_old[x],c_new[x],c_new[y]);
			strcpy(tmpstr,&c_new[x][strlen(c_new[x])-1]);
			sscanf(tmpstr,"%d",&x1); 
			strcpy(tmpstr,&c_new[y][strlen(c_new[y])-1]);
			sscanf(tmpstr,"%d",&y1); 
			if (x1 < y1) {strcpy(c_new[x],c_new[y]);if (diag>0) printf("%s\n",c_new[x]);}
			        else {strcpy(c_new[y],c_new[x]);if (diag>0) printf("%s\n",c_new[y]);}
		}
		actual_chgcnt--;
		c_used[y]=1;
		if (c_trans[y]==1) c_trans[x]=1;
          }
       }
    }

    /* PROJECT-SPECIFIC - do not apply changes for certain module and instance names */

   y=0;
   for (x=0; x<c_cnt; x++)
   {
       if ( (c_used[x] ==0) && (c_trans[x] == 0) &&
          ( (find_string(c_mod[x],"pll")> 0) ||
	    (find_string(c_inst[x],"uu")==1) ||
	    (find_string(c_mod[x],"dly")> 0) ||
	    (find_string(c_mod[x],"sparegates")>0) ) )
       {
		c_used[x]=1;
		if (c_buffers[x]==0) printf("Removed change:  Mod: %s Inst: %s  Old: %s  New: %s\n",c_mod[x],c_inst[x],c_old[x],c_new[x]);
		                else printf("Removed change:  Mod: %s Inst: %s  Old: %s  Buffers %d\n",c_mod[x],c_inst[x],c_old[x],c_buffers[x]);
		actual_chgcnt--;
		y++;
       }
   }
   if (y>0) printf("Removed %d changes to modules with pll, dly, or sparegates in their names, or uu in the instancename\n",y);

   /* count buffers */
   buffer_cnt=0;
   for (x=0;x<c_cnt;x++)
   {
	if (c_used[x]==0) buffer_cnt = buffer_cnt + c_buffers[x];
   }


   printf("Conflicts resolved, %d changes remaining.\n",actual_chgcnt);
   if (buffer_cnt>0) printf("%d buffers will be added\n",buffer_cnt);


   /* next step - write out database if requested */

    if (write_changes_db == 1)
    {
       fprintf(outfile_ptr,"$ START_CHANGES_DATABASE_INPUT %s %s %s %s %d\n",bc_sdf_dly_r,bc_sdf_dly_f,wc_sdf_dly_r,
			      wc_sdf_dly_f,nodenum);

	for (x=0; x< c_cnt; x++)
	{
	  if (c_used[x]==0)
	  {
	    if (strlen(c_mod[x])      > 0) fprintf(outfile_ptr,"%s ",c_mod[x]     ); else fprintf(outfile_ptr,"null ");
	    if (strlen(c_inst[x])     > 0) fprintf(outfile_ptr,"%s ",c_inst[x]    ); else fprintf(outfile_ptr,"null ");
 	    if (strlen(c_old[x])      > 0) fprintf(outfile_ptr,"%s ",c_old[x]     ); else fprintf(outfile_ptr,"null ");
	    if (strlen(c_new[x])      > 0) fprintf(outfile_ptr,"%s ",c_new[x]     ); else fprintf(outfile_ptr,"null ");
	    if (strlen(c_hiername[x]) > 0) fprintf(outfile_ptr,"%s ",c_hiername[x]); else fprintf(outfile_ptr,"null ");
	    if (strlen(c_pin[x])      > 0) fprintf(outfile_ptr,"%s ",c_pin[x]     ); else fprintf(outfile_ptr,"null ");
	    fprintf(outfile_ptr,"%d %d %d\n",c_exact[x],c_trans[x],c_buffers[x]);
	  }
        }
	fprintf(outfile_ptr,"$ END_CHANGES_FILE\n");
	printf("Database file %s written\n",outfile);	    
	if (trpfile_open==1) printf("%s written\n",trptfile);
	fclose(outfile_ptr);
	exit(0);
     }


   /* next step - apply changes */

    printf("\nApplying changes to netlist...\n");
    if (open_sdf==2) open_sdfs();
    total_chgcnt=0;
    fprintf(outfile_ptr,"// IPOed netlist.  Generated by ipofix.c, Rev. 1.1, By Jeff Winston\n");
    fprintf(outfile_ptr,"// Input file: %s, Report file: %s\n",infile,rptfile);
    if (buffer_cnt>0) fprintf(outfile_ptr,"// %d buffers added\n",buffer_cnt);
    fprintf(outfile_ptr,"// %d gates changed.  \n\n\n",actual_chgcnt);

    all_done=0;
    infile_line_cnt=0;
    buffer_cnt=0;
    active_mod=0;
    while (fgets(instr,255,infile_ptr) != NULL)             /* Keep getting lines until no more data */
    {
       ntk = get_tokens(instr);

       if (find_string(tokens[0],"//")==1) ntk=0;
       else
       {
              strcpy(lines[0],instr);
	      infile_line_cnt=1;
       }
       if (ntk > 0) 
       {						/* build full lines out of incomplete ones */
	 done = add_str(instr);
	 while (done == 0) 
	 {
	      if (fgets(instr,255,infile_ptr) == NULL) done=1;
	      else
	      {
	         strcpy(lines[infile_line_cnt],instr);
	         infile_line_cnt++;
		 limit_check(infile_line_cnt,1000,"buffered_lines",instr);
	         done = add_str(instr);
	      }
	 }
	 done=0;
 	 start_line=0;
	 ntk = get_tokens(fullstr);			/* parse the full line */
	 if (strcmp(tokens[0],"module") == 0)
	 {
		if (active_mod==1) 
		{
		     printf("%d changes\n",chgcnt);
		     if (chgcnt>0) fprintf(outfile_ptr,"// ipofix made %d changes to module %s\n\n",chgcnt,modname);
		}
		strcpy(modname,tokens[1]);
		active_mod=0;
		for (x=0; x<c_cnt; x++)
		{
			if (strcmp(c_mod[x],modname)==0)
			{
				printf("Making changes to module %s... ",modname); fflush(stdout);
				active_mod=1;
				chgcnt=0;
				x=c_cnt+10;
			}				
		}		
	 }		
	 else if ((active_mod==1) && (strcmp(tokens[2],"(") ==0)) /* if we're in a module that has changes, and */
	 {							  /* this is an entry - parse it */
		
		if (find_string(tokens[1],"_XQX")>0)	/* get instance name, remove XQX if it exists */
		{
			strncpy(instname,&tokens[1][0],strlen(tokens[1])-4);
			strncpy(&instname[strlen(tokens[1])-4],"\0",1);
		}
		else strcpy(instname,tokens[1]);

		chgnum=0;
		while ((chgnum<c_cnt) && (start_line==0))	/* check changes for instance name */
		{
		    if ((c_used[chgnum]==0) && (strcmp(c_mod[chgnum],modname)==0))
                    {

			if (strcmp(instname,c_inst[chgnum])==0)
			{
				strncpy(tmpstr,&tokens[0][0],strlen(tokens[0])-1);
				strncpy(&tmpstr[strlen(tokens[0])-1],"\0",1);
											/* detect possible problems */

				     if ( (c_exact[chgnum] == 1) && (strcmp(tokens[0],c_old[chgnum])!=0) && (find_string(c_old[chgnum],tmpstr)>0))
				{
					printf("\nWarning: In module %s, instance name %s matches, but old strength\n",
							  modname,c_inst[chgnum]);
					printf("       in input file (%s) doesn't match strength in change database (%s)\n",
						          tokens[0],c_old[chgnum]);
					printf("       Line: %s",lines[0]);	
				}
				else if ( (c_exact[chgnum] == 1) && (strcmp(tokens[0],c_old[chgnum])!=0) && (find_string(c_old[chgnum],tmpstr)==0))
				{
					printf("\nError: In module %s, instance name %s matches, but old type\n",
							  modname,c_inst[chgnum]);
					printf("       in input file (%s) doesn't match type in change database (%s) %s\n",
						          tokens[0],c_old[chgnum]);
					printf("       Line: %s",lines[0]);		
					exit(1);
				}
				else if ((c_exact[chgnum] == 0) && (find_string(c_old[chgnum],tmpstr)==0))
				{
					printf("\nError: In module %s, instance name %s matches, but old generic type\n",
							  modname,c_inst[chgnum]);
					printf("       in input file (%s) doesn't match type in change database (%s)\n",
						          tmpstr,c_old[chgnum]);
					printf("       Line: %s",lines[0]);		
					exit(1);
				}
				else if ((c_exact[chgnum] == -1) && (find_string(c_old[chgnum],tmpstr)==0) && 
				     (strcmp(c_old[chgnum],"null")!=0))
				{
					printf("\nWarning: In module %s, instance name %s matches, but old generic type\n",
							  modname,c_inst[chgnum]);
					printf("       in input file (%s) doesn't match type in change database (%s)\n",
						          tmpstr,c_old[chgnum]);
					printf("       Line: %s",lines[0]);		
					strcpy(c_old[chgnum],tmpstr);
					strcpy(c_new[chgnum],tmpstr);
				}

					/* if we didn't exit on a problem - update the line */
				if (diag>0)  {printf("Updating line %s (bf = %d)\n",lines[0],c_buffers[chgnum]);fflush(stdout);}

				/* for hold buffers - need to add buffer, make SDF entry, and edit entry which has hold */
				/* buffer as input */
				if (c_buffers[chgnum]>0) 
				{
					ntk = get_tokens(fullstr);		/* find nodename to replace on existing line */
					tok=0;
					while ((tok < ntk) && (find_string(tokens[tok],c_pin[chgnum])==0)) tok++;
					if (tok == ntk)
					{
						printf("Error: No %s input %s\n",c_pin[tok],fullstr);
						exit(1);
					}
					tok=tok+2; /* token to replace */
					if (find_string(tokens[tok],"{")>0) tok++;

					for (x=0; x<c_buffers[chgnum]; x++)	/* add hold buffers */
					{
					   if (x == 0) fprintf(outfile_ptr, "  ni01d1 hb%d  ( .i ( %s ), .z ( hn%d ));\n",
								  nodenum,tokens[tok],nodenum);
					   else fprintf(outfile_ptr, "  ni01d1 hb%d  ( .i ( hn%d ), .z ( hn%d ));\n",nodenum,nodenum-1,nodenum);

									       /* add SDF entries */
					   fprintf(bsdf_ptr,"(CELL \n");
					   fprintf(bsdf_ptr,"  (CELLTYPE \"ni01d1\")\n");
					   fprintf(bsdf_ptr,"   (INSTANCE %shb%d )\n",c_hiername[chgnum],nodenum);
					   fprintf(bsdf_ptr,"   (DELAY\n");
					   fprintf(bsdf_ptr,"      (ABSOLUTE\n");
					   fprintf(bsdf_ptr,"         (IOPATH i z  (%s:%s:%s) (%s:%s:%s))\n",bc_sdf_dly_r,
									 bc_sdf_dly_r, bc_sdf_dly_r, bc_sdf_dly_f, bc_sdf_dly_f,
									 bc_sdf_dly_f);
					   fprintf(bsdf_ptr,"      )\n");
					   fprintf(bsdf_ptr,"   )\n");
					   fprintf(bsdf_ptr,"   (DELAY\n");
					   fprintf(bsdf_ptr,"      (ABSOLUTE (PORT i (0.000:0.000:0.000) (0.000:0.000:0.000)))\n");
					   fprintf(bsdf_ptr,")\n");
					   fprintf(bsdf_ptr,")\n");
					   fprintf(wsdf_ptr,"(CELL \n");
					   fprintf(wsdf_ptr,"  (CELLTYPE \"ni01d1\")\n");
					   fprintf(wsdf_ptr,"   (INSTANCE %shb%d )\n",c_hiername[chgnum],nodenum);
					   fprintf(wsdf_ptr,"   (DELAY\n");
					   fprintf(wsdf_ptr,"      (ABSOLUTE\n");
					   fprintf(wsdf_ptr,"         (IOPATH i z  (%s:%s:%s) (%s:%s:%s))\n",wc_sdf_dly_r,
									 wc_sdf_dly_r, wc_sdf_dly_r, wc_sdf_dly_f, wc_sdf_dly_f,
									 wc_sdf_dly_f);
					   fprintf(wsdf_ptr,"      )\n");
					   fprintf(wsdf_ptr,"   )\n");
					   fprintf(wsdf_ptr,"   (DELAY\n");
					   fprintf(wsdf_ptr,"      (ABSOLUTE (PORT i (0.000:0.000:0.000) (0.000:0.000:0.000)))\n");
					   fprintf(wsdf_ptr,")\n");
					   fprintf(wsdf_ptr,")\n");
					   nodenum++;
					}

					y=0; /* y is used for format spacing */
					for (x=0;x<ntk;x++)				/* re-write line */
					{
						if (x==0) fprintf(outfile_ptr,"  ");
						if (y> 78) 
						{
							fprintf(outfile_ptr,"\n     ");
							y=5;
						}		
						if (x !=tok) 
						{
							fprintf(outfile_ptr,"%s ",tokens[x]);
							y=y+strlen(tokens[x])+1;
						}
						else 
						{
							fprintf(outfile_ptr,"hn%d ",nodenum-1);
							y=y+7;
							if (nodenum>=10000) y++;
						}
					}
					if (y > 5) fprintf(outfile_ptr,"\n");

					buffer_cnt=buffer_cnt+c_buffers[chgnum];
					start_line=infile_line_cnt;
				}
				else	/* upsizing a gate... */
				{
					if (diag > 2) printf("New type: %s, new instance %s\n",c_new[chgnum],c_inst[chgnum]);

					fprintf(outfile_ptr,"    %s  %s_XQX ",c_new[chgnum],c_inst[chgnum]);
					ntk = get_tokens(lines[0]);

					for (tok=2;tok<ntk;tok++) fprintf(outfile_ptr,"%s ",tokens[tok]);

					fprintf(outfile_ptr,"\n");
					start_line=1;	
				}
				total_chgcnt++;
				chgcnt++;
				c_used[chgnum] = 1;
			}
                   }
		   chgnum++;
		}
	   }
	   for (x=start_line;x<infile_line_cnt;x++)	/* output remaining lines for gate */
	   	   fprintf(outfile_ptr,"%s",lines[x]);
	   start_line=0;
      }
      else fprintf(outfile_ptr,"%s",instr);
   }
   if (active_mod==1)	/* if we end with an active module */
   {
	printf("%d changes\n",chgcnt);
        if (chgcnt>0) fprintf(outfile_ptr,"// ipofix made %d changes to module %s\n\n",chgcnt,modname);
   }

   /* check for unused changes */
   for (x=0; x<c_cnt; x++)
   {
       if (c_used[x]==0) 
       {
	    /*if (diag>0) */printf("Error: Didn't apply change in module %s: Instance: %s, old: %s  new: %s\n",
			      c_mod[x],c_inst[x],c_old[x],c_new[x]);
       }			      
   }

   /* final info */

   if (total_chgcnt> 0) printf("Made %d changes\n",total_chgcnt);
   if (buffer_cnt>0) printf("Added %d buffers\n",buffer_cnt);
   printf("%s written\n",outfile);
   if (open_sdf==1) 
   {
	fprintf(bsdf_ptr,")\n");
	fprintf(wsdf_ptr,")\n");
	printf("wc.sdf, bc.sdf written\n");
   }
   if (trpfile_open==1) printf("%s written\n",trptfile);
}

/* end of main program */

build_ref_table()	/* builds the module reference table */
{
        modcnt=0; i_cnt=0;
        printf("\nBuilding reference table...\n");

        while (fgets(instr,255,infile_ptr) != NULL)             /* Keep getting lines until no more data */
        {
            ntk = get_tokens(instr);
            if (ntk > 0) 
	    {						/* build full lines out of incomplete ones */
	       done = add_str(instr);
	       while (done == 0) 
	       {
	          if (fgets(instr,255,infile_ptr) == NULL) done=1;
		  else done = add_str(instr);
	       }
	       done=0;
	       ntk = get_tokens(fullstr);			/* parse the full line */

	       if (strcmp(tokens[0],"module") == 0)
	       {
		   strcpy(i_mods[modcnt],tokens[1]);
		   printf("Reading module %s\n",tokens[1]);fflush(stdout);
		   modcnt++;		
		   limit_check(modcnt,MAX_MODULES,"modules_in_input_file",instr);
	       }
	       else if (strcmp(tokens[2],"(")==0)
	       {
		   for (x=0; x<(modcnt-1);x++)
		   {
			if (strcmp(i_mods[x],tokens[0])==0)
			{
				strcpy(i_mmod[i_cnt], i_mods[modcnt-1]);
				strcpy(i_cmod[i_cnt], tokens[0]);
				strcpy(i_inst[i_cnt], tokens[1]);
				i_cnt++;
				x = modcnt;
			}
		   }
	      }
          }
       }
        fclose(infile_ptr);
}

read_reference_table()	/* reads in a pre-built, previously written out module reference table  */
{
   i_cnt=0;
   printf("Reading in reference table\n");

   while (fgets(instr,255,dbfile_ptr) != NULL)            
   {
	ntk = get_tokens(instr);
	if (ntk > 1)
	{
		strcpy(i_mmod[i_cnt],tokens[0]);
		strcpy(i_cmod[i_cnt],tokens[1]);
		strcpy(i_inst[i_cnt],tokens[2]);
		i_cnt++;   
		limit_check(i_cnt,MAX_MOD_CALLS,"module_calls",instr);
	}
   }
   printf("%d module references.\n",i_cnt);
}

int add_str(str)	/* used for constructing full verilog statements out of multiple lines */
char *str;
{
	char tmpstr[65536];
	long loc, len, sp,w;
	len = strlen(str);
	if (all_done == 1)
        {
	   strcpy(fullstr,"\0");
  	   all_done=0;
	}
	if (len>0) len--;
	for (loc=0; loc < len; loc++)
	{
		if ( (parsing_comment==1) && (str[loc]=='*') && (str[loc+1]=='/') )
		{
		  parsing_comment=0;
		  loc++;
		}		  
		else if (parsing_comment==0)
		{
		         if ((str[loc]=='/') && (str[loc+1]=='*'))  parsing_comment=1;
		    else if ((str[loc]=='/') && (str[loc+1]=='/'))  loc = len;
		    else if (str[loc]=='(')   strncat(fullstr," ( ",3); 
		    else if (str[loc]==')')   strncat(fullstr," ) ",3); 
		    else if (str[loc]=='{')   strncat(fullstr," { ",3); 
		    else if (str[loc]=='}')   strncat(fullstr," } ",3); 
		    else if (str[loc]==';') 
		    {
		        strncat(fullstr,&str[loc],1); 
		        strncat(fullstr,"\0",1);
		        all_done=1;
                    }			       
		    else if  (
			 (loc+8 < len)     &&
			 (str[loc+0]=='e') &&
			 (str[loc+1]=='n') &&
			 (str[loc+2]=='d') &&
			 (str[loc+3]=='m') &&
			 (str[loc+4]=='o') &&
			 (str[loc+5]=='d') &&
			 (str[loc+6]=='u') &&
			 (str[loc+7]=='l') &&
			 (str[loc+8]=='e') )
		    {
		        for (w=loc;w<loc+9;w++)
		        {
		            strncat(fullstr,&str[w],1); 
		        }
		        strncat(fullstr,"\0",1);
		        all_done=1;
                    }			       
		    else  strncat(fullstr,&str[loc],1); 
	       }	        
	}	        
	if (all_done==1)
	{
	    sp=0;
	    strcpy(tmpstr,fullstr);
	    strcpy(fullstr,"\0");
	    len = strlen(tmpstr);
	    for (w=0;w<len;w++)
	    {
		if ((tmpstr[w] != ' ') || (sp==0))
		{
		    strncat(fullstr,&tmpstr[w],1); 
		    if (tmpstr[w] == ' ') sp=1; else sp=0;
		}
	    }
 	    strncat(fullstr,"\0",1);
	}
	return(all_done);
}



int get_tokens(str)           /* get tokens on line. Put in global tokens[32][32], return  count */
char *str;
{
   int len, loc, cnt, sp, start;

   len = strlen(str);  sp=1; cnt=0; 

   for (loc=0; loc <= len; loc++)
   {
     
       if ( (loc < len) && (str[loc] != ' ') && (str[loc] != '\n') && (str[loc] != '	') && (str[loc] != "\0") )
       {
           if (sp==1)		/* if state = waiting */
           {
               cnt++;		/* incr token count, flip state to running, assign start */

	       limit_check(cnt,MAXTOKENS,"tokens_on_line",str);
               sp=0;
               start=loc;
           }
       }
       else		/* if not valid character */
       {
           if (sp==0)	/* if state = running, flip state to waiting, append good characters to string */
           {
               sp=1;
               strncpy(tokens[cnt-1],&str[start],loc-start);
               strncpy(&tokens[cnt-1][loc-start],"\0",1);

           }
       }
   }
   return(cnt);
}
int get_tokens_hier(str)           /* get tokens on line. Put in global tokens[32][32], return  count */
char *str;			   /* differs from above by selection of delimiters */
{
   int len, loc, cnt, sp, start;

   len = strlen(str);  sp=1; cnt=0; 

   for (loc=0; loc <= len; loc++)
   {
     
       if ( (loc < len) && (str[loc] != ' ') && (str[loc] != '\n') && (str[loc] != '/') && (str[loc] != "\0") )
       {
           if (sp==1)		/* if state = waiting */
           {
               cnt++;		/* incr token count, flip state to running, assign start */
	       limit_check(cnt,MAX_HIERARCHY,"levels_of_hierarchy",str);
               sp=0;
               start=loc;
           }
       }
       else		/* if not valid character */
       {
           if (sp==0)	/* if state = running, flip state to waiting, append good characters to string */
           {
               sp=1;
               strncpy(itokens[cnt-1],&str[start],loc-start);
               strncpy(&itokens[cnt-1][loc-start],"\0",1);

           }
       }
   }
   return(cnt);
}
int get_par_tokens(str,n)           /* get tokens on line. Put in global tokens[32][32], return  count */
char *str; long n;		    /* differs from above by selection of delimiters */
{
   int len, loc, cnt, sp, start;

   len = strlen(str);  sp=1; cnt=n; 

   for (loc=0; loc <= len; loc++)
   {
     
       if ( (loc < len) && (str[loc] != ' ') && (str[loc] != '\n') && (str[loc] != '	') && 
                           (str[loc] != '(') && (str[loc] != ')') && (str[loc] != "\0") )
       {
           if (sp==1)		/* if state = waiting */
           {
               cnt++;		/* incr token count, flip state to running, assign start */
	       limit_check(cnt,MAXTOKENS,"tokens_on_line",str);
               sp=0;
               start=loc;
           }
       }
       else		/* if not valid character */
       {
           if (sp==0)	/* if state = running, flip state to waiting, append good characters to string */
           {
               sp=1;
               strncpy(tokens[cnt-1],&str[start],loc-start);
               strncpy(&tokens[cnt-1][loc-start],"\0",1);

           }
       }
   }
   return(cnt);
}
long find_string(field, source)         /* returns LOC of start of source in field, 0= not found */
char *field;
char *source;
{
    long flen;
    long slen; 
    char    *sPtr;
    long loc;

    flen = strlen(field);
    slen = strlen(source);
    if (slen > flen) return (0);     /* error */
    else
    {
        for (loc=0, sPtr = field; loc <= flen-slen; loc++, sPtr++)
        {
            if (strncmp(source,sPtr, slen) ==0) return(loc+1);
        }
    }
    return (0);
}

long get_rpt_line(lines)	/* get next interesting line in report file */
long lines;
{
    long x;
    for (x=0; x<lines; x++)
    {
        if (fgets(instr,255,rptfile_ptr) == NULL) return(0);
	else 
	{
	   rpt_line++;
        }
    }
    return(1);
}


/* resolves a hierarchical reference to a module using the module reference table */
get_inst_and_pin(modname,hier_instname,type)  /* returns intk, contents of itokens, and modtmp */
char *modname; char *hier_instname; long type;
{
	long x; long done;
	intk = get_tokens_hier(hier_instname);
	strcpy(modtmp,modname);
	ilevel=0;
	while (intk >(2-type))
	{
	 	done=0;
		for (x=0; x<i_cnt;x++)
		{
			if ((strcmp(i_mmod[x],modtmp)==0) && (strcmp(i_inst[x],itokens[ilevel])==0)) 
			{
				if (diag>0) 
				printf("For module %s, found that instance name %s points to module %s\n",
						    modtmp,itokens[ilevel],i_cmod[x]);

				strcpy(modtmp,i_cmod[x]);
				x = i_cnt;
				ilevel++;					
				done=1;
				intk--;
			}
		}		
		if (done==0)
		{
		        printf("Error: For module %s, couldn't find module name for instance %s in line %s\n",
					    modtmp,itokens[ilevel], instr);
			exit(1);
		}
	}
	inst_tok = ilevel;
	pin_tok = inst_tok+1;
}

put_hiername()		/* construct hierarchical name from itokens */
{
     long x;
     for (x=0;x<inst_tok;x++) 
     {
	strncat(c_hiername[c_cnt],&itokens[x],strlen(itokens[x]));
	strncat(c_hiername[c_cnt],"/",1);
     }
     strncat(c_hiername[c_cnt],"\0",1);
}

limit_check(value,limit,errstr,filestr)		/* array limit check, error reporter */
long value, limit; char *errstr; char *filestr;
{
	if (value >limit -1)
	{
		printf("Error: Size of arrays holding <%s> exceeded.\nLast line processed: %s\nExiting...\n",
		errstr,filestr);
		exit(1);
	}
}

trans_entry(type,name,old,trans)	/* write entry to transitions report file */
long type; char * name; char * old; float trans;
{
    long x, done;

    done=0;
    for (x=0; x<t_cnt; x++)
    {
	if ((t_type[x] == type) && (strcmp(t_name[x],name)==0))
	{
		done=1;
		x=t_cnt;
	}
     }
     if (done == 0)
     {
	     if  (type == TRANS_CANT_FIX)           fprintf(trptfile_ptr,"Can't Fix:      %48-s old: %8-s transition: %f\n",name,old,trans);
	else if  (type == TRANS_SKIP)               fprintf(trptfile_ptr,"Skipped Fix:    %48-s old: %8-s transition: %f\n",name,old,trans);
	else if  (type == TRANS_NONEED)             fprintf(trptfile_ptr,"No Need to Fix: %48-s old: %8-s transition: %f\n",name,old,trans);
	else if ((type == TRANS_FIXED) && (diag>0)) fprintf(trptfile_ptr,"Made Fix:       %48-s old: %8-s transition: %f\n",name,old,trans);
	t_type[t_cnt] = type;
	strcpy(t_name[t_cnt],name);
	t_cnt++;
	limit_check(t_cnt,MAX_TRANSITIONS,"stored_transitions",instr);
     }
}


ck_argv(argc)	/* print command line help if wrong # of arguments */
int argc;
{
    argcnt++; 
    if (argcnt > argc)
    {

        printf("\nError:  Incorrect number of parameters on command line\n");
        printf("Usage: ipofix [< -h|-t|-d|-c >] <input filename> [<report filename>] <output filename> [<database filename>] \n");
        printf("	Where -h prints help and exits  (use of | more recommended) \n");
        printf("	      -t generates a database to <output filename> and exits \n");
        printf(" 	      -d enables diagnostic mode \n");
        printf("	      -c enables normal processing until all the changes are processed, then \n");
        printf("                 the program writes the change database to the output file in tabular form and exits. \n\n");
        exit(0);
    }
}

open_trans_file()	/* open the transition report file */
{
	if (trpfile_open==0)
	{
	       if ( (trptfile_ptr=fopen(trptfile, "w")) ==NULL )
	       {
	               printf("Error:  Error opening input file /%s/\n",trptfile);
		       exit(1);
	       }
	       trpfile_open=1;
	       printf("Opening file %s for transition errors\n",trptfile);
	}
}

open_sdfs()	/* file open routine for SDfs */
{
    open_sdf=1;
    if ((bsdf_ptr=fopen(bsdf_file, "w")) ==NULL)
    {
        printf("Error:  Error opening output file /%s/\n\n",bsdf_file);
        exit(1);
    }
    if ((wsdf_ptr=fopen(wsdf_file, "w")) ==NULL)
    {
        printf("Error:  Error opening output file /%s/\n\n",wsdf_file);
        exit(1);
    }

	fprintf(bsdf_ptr,"(DELAYFILE\n");
	fprintf(bsdf_ptr,"   (SDFVERSION \"2.1\")\n");
	fprintf(bsdf_ptr,"   (DESIGN \"ipofix_output\")\n");
	fprintf(bsdf_ptr,"   (DATE \"00/00/00 00:00:00\")\n");
	fprintf(bsdf_ptr,"   (VENDOR \"VLSI Technology Inc.\")\n");
	fprintf(bsdf_ptr,"   (PROGRAM \"ipofix\")\n");
	fprintf(bsdf_ptr,"   (VERSION \"v0\")\n");
	fprintf(bsdf_ptr,"   (DIVIDER /)\n");
	fprintf(bsdf_ptr,"   (VOLTAGE 1.890:1.890:1.890)\n");
	fprintf(bsdf_ptr,"   (PROCESS \"FAST:FAST:FAST\")\n");
	fprintf(bsdf_ptr,"   (TEMPERATURE -40.000:-40.000:-40.000)\n");
	fprintf(bsdf_ptr,"   (TIMESCALE   1 ns)\n");

	fprintf(wsdf_ptr,"(DELAYFILE\n");
	fprintf(wsdf_ptr,"   (SDFVERSION \"2.1\")\n");
	fprintf(wsdf_ptr,"   (DESIGN \"ipofix_output\")\n");
	fprintf(wsdf_ptr,"   (DATE \"00/00/00 00:00:00\")\n");
	fprintf(wsdf_ptr,"   (VENDOR \"VLSI Technology Inc.\")\n");
	fprintf(wsdf_ptr,"   (PROGRAM \"ipofix\")\n");
	fprintf(wsdf_ptr,"   (VERSION \"v0\")\n");
	fprintf(wsdf_ptr,"   (DIVIDER /)\n");
	fprintf(wsdf_ptr,"   (VOLTAGE 1.700:1.700:1.700)\n");
	fprintf(wsdf_ptr,"   (PROCESS \"SLOW:SLOW:SLOW\")\n");
	fprintf(wsdf_ptr,"   (TEMPERATURE 110.000:110.000:110.000)\n");
	fprintf(wsdf_ptr,"   (TIMESCALE   1 ns)\n");

}

